# This file is part of GNOME Shell Volume Mixer
# Copyright (C) 2021 Alexander Hofbauer <alex@derhofbauer.at>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .pulseaudio import Pulseaudio
from . import log
from . import libpulse


class Cards(Pulseaudio):
    def build_callback(self):
        return libpulse.card_info_cb_t(self.pa_cb)

    def get_by_index(self, index, callback):
        return libpulse.context_get_card_info_by_index(self._context, index, callback, None)

    def get_by_name(self, name, callback):
        return libpulse.context_get_card_info_by_name(self._context, name, callback, None)

    def get_all(self, callback):
        return libpulse.context_get_card_info_list(self._context, callback, None)

    def cb_data(self, pa_card):
        card_name = pa_card.name.decode('utf8')

        try:
            alsa_card = libpulse.proplist_gets(pa_card.proplist, b'alsa.card')
            alsa_card = int(alsa_card.decode('utf8')) if alsa_card else None
        except Exception:
            log.debug('No property "alsa.card"')
            alsa_card = None

        try:
            description = libpulse.proplist_gets(pa_card.proplist, b'device.description')
            description = description.decode('utf8') if description else None
        except Exception:
            log.debug('No property "device.description"')
            description = None

        if not description:
            try:
                description = libpulse.proplist_gets(pa_card.proplist, b'alsa.card_name')
                description = description.decode('utf8') if description else None
            except Exception:
                log.debug('No property "alsa.card_name"')
                description = card_name

        card = {
            'index': pa_card.index,
            'alsaCard': alsa_card,
            'name': card_name,
            'description': description,
            'active_profile': None,
            'profiles': {
            },
            'ports': {
            },
        }

        if pa_card.active_profile and pa_card.active_profile[0]:
            ap = pa_card.active_profile[0]
            card['active_profile'] = ap.name.decode('utf8')

        for i in range(0, pa_card.n_profiles):
            if not pa_card.profiles2[i] or not pa_card.profiles2[i][0]:
                continue

            profile = pa_card.profiles2[i][0]
            name = profile.name.decode('utf8')

            card['profiles'][name] = {
                'name': name,
                'description': profile.description.decode('utf8'),
                'available': bool(profile.available),
            }

        for index in range(0, pa_card.n_ports):
            if not pa_card.ports[index] or not pa_card.ports[index][0]:
                continue

            port = pa_card.ports[index][0]
            name = port.name.decode('utf8')

            card['ports'][name] = {
                'name': name,
                'description': port.description.decode('utf8'),
                'direction': 'out' if port.direction == 1 else 'in',
                'available': True if port.available == 2 else (False if port.available == 1 else None),
            }

        return card



#!/usr/bin/env python3
#
# Usage: query.py [cards|sinks] [index or name, omit for all data]
#
# Output is either a JSON object or an array of all data available, depending on
# whether an index / name was passed or no parameters at all.
#
# This file is part of GNOME Shell Volume Mixer
# Copyright (C) 2021 Alexander Hofbauer <alex@derhofbauer.at>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import sys

from lib import log
from lib.cards import Cards
from lib.sinks import Sinks
from lib.libpulse import NULL_ID

if len(sys.argv) < 2:
    print('Need a type to query')
    sys.exit(1)

op_type = sys.argv[1]
index = name = None

if len(sys.argv) > 2:
    filter_arg = sys.argv[2]
    if filter_arg.isdigit():
        index = int(filter_arg)
    else:
        name = filter_arg


if index == NULL_ID:
    result = {}

elif op_type == 'cards':
    with Cards() as cards:
        result = cards.get_info(index=index, name=name)

elif op_type == 'sinks':
    with Sinks() as sinks:
        result = sinks.get_info(index=index, name=name)

else:
    print('Invalid type', op_type, 'requested')
    sys.exit(1)


print(json.dumps(result, indent=4 if log.DEBUG else None))



# This file is part of GNOME Shell Volume Mixer
# Copyright (C) 2021 Alexander Hofbauer <alex@derhofbauer.at>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .pulseaudio import Pulseaudio
from . import log
from . import libpulse


class Sinks(Pulseaudio):
    def build_callback(self):
        return libpulse.sink_info_cb_t(self.pa_cb)

    def get_by_index(self, index, callback):
        return libpulse.context_get_sink_info_by_index(self._context, index, callback, None)

    def get_by_name(self, name, callback):
        return libpulse.context_get_sink_info_by_name(self._context, name, callback, None)

    def get_all(self, callback):
        return libpulse.context_get_sink_info_list(self._context, callback, None)

    def cb_data(self, pa_sink):
        sink_name = pa_sink.name.decode('utf8')
        description = pa_sink.description.decode('utf8')

        if not description:
            try:
                description = libpulse.proplist_gets(pa_sink.proplist, b'device.description')
                description = description.decode('utf8') if description else None
            except Exception:
                log.debug('No property "device.description"')
                description = sink_name

        try:
            alsa_card = libpulse.proplist_gets(pa_sink.proplist, b'alsa.card')
            alsa_card = int(alsa_card.decode('utf8')) if alsa_card else None
        except Exception:
            log.debug('No property "alsa.card"')
            alsa_card = None

        sink = {
            'index': pa_sink.index,
            'alsaCard': alsa_card,
            'name': sink_name,
            'description': description,
            'card': pa_sink.card if pa_sink.card != libpulse.NULL_ID else None,
            'active_port': None,
            'ports': {
            },
        }

        if pa_sink.active_port and pa_sink.active_port[0]:
            ap = pa_sink.active_port[0]
            sink['active_port'] = ap.name.decode('utf8')

        for index in range(0, pa_sink.n_ports):
            if not pa_sink.ports[index] or not pa_sink.ports[index][0]:
                continue

            port = pa_sink.ports[index][0]
            name = port.name.decode('utf8')

            sink['ports'][name] = {
                'name': name,
                'description': port.description.decode('utf8'),
                'type': port.type,
                'available': True if port.available == 2 else (False if port.available == 1 else None),
            }

        return sink



# This file is part of GNOME Shell Volume Mixer
# Copyright (C) 2021 Alexander Hofbauer <alex@derhofbauer.at>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from time import sleep

from . import log
from . import libpulse


class Pulseaudio:
    _pa_state = libpulse.CONTEXT_UNCONNECTED
    _op_done = False

    _iterations_max = 10000
    _iterations_intrv = 0.0001


    _data = {}
    _error = {
        'success': False,
        'error': None,
    }

    def __init__(self):
        self._pa_mainloop = libpulse.mainloop_new()
        self._pa_mainloop_api = libpulse.mainloop_get_api(self._pa_mainloop)

        self._context = libpulse.context_new(self._pa_mainloop_api, b'ShellVolumeMixer')
        self._context_notify_cb = libpulse.context_notify_cb_t(self.context_notify_cb)

        libpulse.context_set_state_callback(self._context, self._context_notify_cb, None)

    def __enter__(self):
        libpulse.context_connect(self._context, None, 0, None)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        libpulse.context_disconnect(self._context)
        libpulse.context_unref(self._context)
        libpulse.mainloop_free(self._pa_mainloop)

    def context_notify_cb(self, context, userdata):
        try:
            self._pa_state = libpulse.context_get_state(context)
        except Exception:
            self._pa_state = libpulse.CONTEXT_FAILED
            log.debug('Context failed')

    def get_info(self, index=None, name=None):
        log.debug('Querying details...')
        operation = None
        count = 0

        while self._pa_state != libpulse.CONTEXT_TERMINATED:
            if self._op_done:
                break

            if self._pa_state == libpulse.CONTEXT_FAILED:
                self._data = self._error
                self._data['error'] = 'context failed'
                break

            if count >= self._iterations_max:
                log.debug(f'Stopping iterations after {self._iterations_intrv * self._iterations_max}s (bug?)')
                break

            if count > 0 and count % (self._iterations_max / 2) == 0:
                log.debug(f'Delaying after {self._iterations_intrv * count}s (bug?)')
                sleep(2)

            if self._pa_state == libpulse.CONTEXT_READY and operation is None:
                callback = self.build_callback()

                if name:
                    log.debug('Requesting details for', name)
                    operation = self.get_by_name(name.encode('utf8'), callback)

                elif index and index >= 0:
                    log.debug('Requesting details for', index)
                    operation = self.get_by_index(index, callback)

                else:
                    log.debug('Requesting all available data')
                    operation = self.get_all(callback)

            libpulse.mainloop_iterate(self._pa_mainloop, 0, None)
            count += 1
            sleep(self._iterations_intrv)

        if operation:
            libpulse.operation_unref(operation)

        log.debug('Query done')

        return self._data

    def pa_cb(self, context, struct, eol, user_data):
        log.debug('In callback')

        if eol:
            self._op_done = True
            log.debug('All done')
            return

        if not struct or not struct[0]:
            log.debug('No data received for callback')
            return

        item = self.cb_data(struct[0])

        if item and 'index' in item:
            self._data[item['index']] = item

        log.debug('Callback done')

    @abc.abstractmethod
    def build_callback(self):
        return

    @abc.abstractmethod
    def get_by_index(self, index, callback):
        return

    @abc.abstractmethod
    def get_by_name(self, name, callback):
        return

    @abc.abstractmethod
    def get_all(self, callback):
        return

    @abc.abstractmethod
    def cb_data(self, data):
        return None



# This library was generated through introspection of
#     /usr/include/pulse/introspect.h and
#     /usr/include/pulse/mainloop.h.
#     (E.g.:
#         python /usr/share/pyshared/ctypeslib/h2xml.py /usr/include/pulse/introspect.h -o pa.xml
#         python /usr/share/pyshared/ctypeslib/xml2py.py pa.xml -k f -l /usr/lib/libpulse.so -o pa.py
#     )
#
# License according to the header files listed above:
# GNU Lesser General Public License
#
# This file is part of GNOME Shell Volume Mixer
# Copyright (C) 2014 Alexander Hofbauer <alex@derhofbauer.at>


from ctypes import *

try:
    lib = CDLL('libpulse.so.0')
except:
    lib = CDLL('libpulse.so')

STRING = c_char_p
WSTRING = c_wchar_p

class mainloop(Structure):
    pass

class mainloop_api(Structure):
    pass

class spawn_api(Structure):
    pass

class context(Structure):
    pass

class operation(Structure):
    pass

class proplist(Structure):
    pass

class card_profile_info(Structure):
    _fields_ = [
        ('name', STRING),
        ('description', STRING),
        ('n_sinks', c_uint32),
        ('n_sources', c_uint32),
        ('priority', c_uint32),
    ]

class card_profile_info2(Structure):
    _fields_ = [
        ('name', STRING),
        ('description', STRING),
        ('n_sinks', c_uint32),
        ('n_sources', c_uint32),
        ('priority', c_uint32),
        ('available', c_int),
    ]

class card_port_info(Structure):
    _fields_ = [
        ('name', STRING),
        ('description', STRING),
        ('priority', c_uint32),
        ('available', c_int),
        ('direction', c_int),
        ('n_profiles', c_uint32),
        ('profiles', POINTER(POINTER(card_profile_info))),
        ('proplist', POINTER(proplist)),
        ('latency_offset', c_int64),
        ('profiles2', POINTER(POINTER(card_profile_info2))),
    ]

class card_info(Structure):
    _fields_ = [
        ('index', c_uint32),
        ('name', STRING),
        ('owner_module', c_uint32),
        ('driver', STRING),
        ('n_profiles', c_uint32),
        ('profiles', POINTER(card_profile_info)),
        ('active_profile', POINTER(card_profile_info)),
        ('proplist', POINTER(proplist)),
        ('n_ports', c_uint32),
        ('ports', POINTER(POINTER(card_port_info))),
        ('profiles2', POINTER(POINTER(card_profile_info2))),
        ('active_profile2', POINTER(card_profile_info2)),
    ]

class sink_port_info(Structure):
    _fields_ = [
        ('name', STRING),
        ('description', STRING),
        ('priority', c_uint32),
        ('available', c_int),
        ('availability_group', STRING),
        ('type', c_uint32),
    ]

class format_info(Structure):
    _fields_ = [
        ('encoding', c_int),
        ('plist', POINTER(proplist)),
    ]

class sample_spec(Structure):
    _fields_ = [
        ('format', c_int),
        ('rate', c_uint32),
        ('channels', c_uint8),
    ]

class pa_channel_map(Structure):
    _fields_ = [
        ('channels', c_uint8),
        ('map', c_int * int(32)),
    ]

class pa_cvolume(Structure):
    _fields_ = [
        ('channels', c_uint8),
        ('values', c_uint32 * int(32)),
    ]

class sink_info(Structure):
    _fields_ = [
        ('name', STRING),
        ('index', c_uint32),
        ('description', STRING),
        ('sample_spec', sample_spec),
        ('channel_map', pa_channel_map),
        ('owner_module', c_uint32),
        ('volume', pa_cvolume),
        ('mute', c_int),
        ('monitor_source', c_uint32),
        ('monitor_source_name', STRING),
        ('latency', c_uint64),
        ('driver', STRING),
        ('flags', c_int),
        ('proplist', POINTER(proplist)),
        ('configured_latency', c_uint64),
        ('base_volume', c_uint32),
        ('state', c_int),
        ('n_volume_steps', c_uint32),
        ('card', c_uint32),
        ('n_ports', c_uint32),
        ('ports', POINTER(POINTER(sink_port_info))),
        ('active_port', POINTER(sink_port_info)),
        ('n_formats', c_uint8),
        ('formats', POINTER(POINTER(format_info))),
    ]

mainloop_new = lib.pa_mainloop_new
mainloop_new.restype = POINTER(mainloop)
mainloop_new.argtypes = []
mainloop_get_api = lib.pa_mainloop_get_api
mainloop_get_api.restype = POINTER(mainloop_api)
mainloop_get_api.argtypes = [POINTER(mainloop)]
mainloop_iterate = lib.pa_mainloop_iterate
mainloop_iterate.restype = c_int
mainloop_iterate.argtypes = [POINTER(mainloop), c_int, POINTER(c_int)]
mainloop_free = lib.pa_mainloop_free
mainloop_free.restype = None
mainloop_free.argtypes = [POINTER(mainloop)]

context_flags = c_int  # enum
context_flags_t = context_flags

context_state = c_int  # enum
context_state_t = context_state
# values for enumeration 'context_state'
CONTEXT_UNCONNECTED = 0
CONTEXT_CONNECTING = 1
CONTEXT_AUTHORIZING = 2
CONTEXT_SETTING_NAME = 3
CONTEXT_READY = 4
CONTEXT_FAILED = 5
CONTEXT_TERMINATED = 6

context_new = lib.pa_context_new
context_new.restype = POINTER(context)
context_new.argtypes = [POINTER(mainloop_api), STRING]
context_notify_cb_t = CFUNCTYPE(None, POINTER(context), c_void_p)
context_set_state_callback = lib.pa_context_set_state_callback
context_set_state_callback.restype = None
context_set_state_callback.argtypes = [POINTER(context), context_notify_cb_t, c_void_p]
context_connect = lib.pa_context_connect
context_connect.restype = c_int
context_connect.argtypes = [POINTER(context), STRING, context_flags_t, POINTER(spawn_api)]
context_disconnect = lib.pa_context_disconnect
context_disconnect.restype = None
context_disconnect.argtypes = [POINTER(context)]
context_unref = lib.pa_context_unref
context_unref.restype = None
context_unref.argtypes = [POINTER(context)]
context_get_state = lib.pa_context_get_state
context_get_state.restype = context_state_t
context_get_state.argtypes = [POINTER(context)]

operation_unref = lib.pa_operation_unref
operation_unref.restype = None
operation_unref.argtypes = [POINTER(operation)]

card_info_cb_t = CFUNCTYPE(None, POINTER(context), POINTER(card_info), c_int, c_void_p)
context_get_card_info_by_index = lib.pa_context_get_card_info_by_index
context_get_card_info_by_index.restype = POINTER(operation)
context_get_card_info_by_index.argtypes = [POINTER(context), c_uint32, card_info_cb_t, c_void_p]
context_get_card_info_by_name = lib.pa_context_get_card_info_by_name
context_get_card_info_by_name.restype = POINTER(operation)
context_get_card_info_by_name.argtypes = [POINTER(context), STRING, card_info_cb_t, c_void_p]
context_get_card_info_list = lib.pa_context_get_card_info_list
context_get_card_info_list.restype = POINTER(operation)
context_get_card_info_list.argtypes = [POINTER(context), card_info_cb_t, c_void_p]

sink_info_cb_t = CFUNCTYPE(None, POINTER(context), POINTER(sink_info), c_int, c_void_p)
context_get_sink_info_by_index = lib.pa_context_get_sink_info_by_index
context_get_sink_info_by_index.restype = POINTER(operation)
context_get_sink_info_by_index.argtypes = [POINTER(context), c_uint32, sink_info_cb_t, c_void_p]
context_get_sink_info_by_name = lib.pa_context_get_sink_info_by_name
context_get_sink_info_by_name.restype = POINTER(operation)
context_get_sink_info_by_name.argtypes = [POINTER(context), STRING, sink_info_cb_t, c_void_p]
context_get_sink_info_list = lib.pa_context_get_sink_info_list
context_get_sink_info_list.restype = POINTER(operation)
context_get_sink_info_list.argtypes = [POINTER(context), sink_info_cb_t, c_void_p]

proplist_gets = lib.pa_proplist_gets
proplist_gets.restype = STRING
proplist_gets.argtypes = [POINTER(proplist), STRING]

proplist_to_string = lib.pa_proplist_to_string
proplist_to_string.restype = STRING
proplist_to_string.argtypes = [POINTER(proplist)]

# this is a "magic" number (probably -1 at int32, but unsigned?) and tells us there's something fishy
NULL_ID = 4294967295
